__all__ = [
    "shuffle",
    "dc_removal",
    "add_noise",
    "random_scaling",
    "random_overdrive",
    "random_lowpass",
    "random_highpass",
    "random_peak_eq",
]


# Augmentations support only Tensor, not numpy arrays

import torchaudio.functional as F
from torch import Tensor
from typing import Tuple
import numpy as np
import torch


from torch_utilities.audio import invert_db, rms


def shuffle(x: Tensor) -> Tensor:
    """
    Shuffles a tensor over the batches.

    Parameters
    ----------
    x : Tensor
        Target tensor of shape (B, ...)

    Returns
    -------
    Tensor
        Shuffled tensor, same shape of x
    """
    y = x[torch.randperm(x.shape[0])]
    return y


# L-a-m-b-d-a-s = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

_db_to_lin_range = lambda xs: [invert_db(x) for x in xs]

_rand_in_range = lambda a, b, n, dev: torch.rand(n, device=dev) * (b - a) + a

_1rand_in_range = lambda a, b: np.random.rand(1)[0] * (b - a) + a

_max_over_batch = lambda x: torch.max(x.flatten(1), dim=1)[0]

_normalize = lambda x: x / _max_over_batch(x + 1e-12)[:, None, None]

_expand2 = lambda x: x[:, None]

_expand3 = lambda x: x[:, None, None]

_safe_div = lambda n, d, eps: n / (d + eps)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


def dc_removal(x: Tensor) -> Tensor:
    """
    DC removal for each batch and channel.

    Parameters
    ----------
    x : Tensor
        Input signal of shape (B, C, T)

    Returns
    -------
    Tensor
        Processed output without DC removal of shape (B, C, T)
    """
    m = x.mean(dim=2)
    x -= m[..., None]
    return x


def add_noise(
    x: Tensor, n: Tensor, snr_range: Tuple[float, float]
) -> Tuple[Tensor, Tensor]:
    """
    Add noise at various SNRs.

    Parameters
    ----------
    x : Tensor
        Target tensor of shape (B, C, T)
    n : Tensor
        Noise tensor of shape (B, C, T)
    snr_range : Tuple[float, float]
        Max and min SNR

    Returns
    -------
    Tuple[Tensor]
        (
            noisy version of x of shape (B, C, T),
            per-batch SNR of shape (B,)
    """
    a, b = _db_to_lin_range(snr_range)
    snr = _rand_in_range(a, b, x.shape[0], x.device)

    # the output will match these peaks
    x_peaks = _max_over_batch(x)

    # scaling the noise
    eps = 1e-8
    n = _safe_div(n, _expand2(rms(n)), eps)
    n *= _expand2(rms(x))
    n = _safe_div(n, _expand3(snr), eps)

    # summing and scaling back
    y = x + n
    y = _normalize(y)
    y *= _expand3(x_peaks)

    return y, snr


def random_scaling(x: Tensor, range_db: Tuple[float, float]) -> Tuple[Tensor, Tensor]:
    """
    Applies a random scaling over the batches

    Parameters
    ----------
    x : Tensor
        Target tensor of shape (B, ...)
    range_db : Tuple[float, float]
        Range of the scaling in dB

    Returns
    -------
    Tuple[Tensor, Tensor]
        (
            scaled version of x of shape (B, ...),
            per-batch scaling of shape (B,)
        )
    """
    a, b = _db_to_lin_range(range_db)
    scale = _rand_in_range(a, b, x.shape[0], x.device)
    x = _normalize(x)
    x *= _expand3(scale)
    return x, scale


def random_overdrive(
    x: Tensor,
    gain_range: Tuple[float, float] = (60, 100),
    colour_range: Tuple[float, float] = (0, 100),
) -> Tensor:
    """
    Applies an overdrive to the input.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (..., T)
    gain_range : Tuple[float, float], optional
        Range of the gain parameter, by default (60, 100)
    colour_range : Tuple[float, float], optional
        Range of the color parameter, by default (0, 100)

    Returns
    -------
    Tensor
        Distorted version of x of shape (..., T)
    """

    x_peaks = _max_over_batch(x)

    gain = _1rand_in_range(*gain_range)
    color = _1rand_in_range(*colour_range)
    y = F.overdrive(x, gain, color)

    y = _normalize(y)
    y *= _expand3(x_peaks)

    return y


def random_lowpass(
    x: Tensor,
    sample_rate: int,
    cutoff_range: Tuple[float, float],
    q_range: Tuple[float, float],
) -> Tensor:
    """
    Apply a random lowpass filter to the input batch.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (..., T)
    sample_rate : int
        Sample rate of the input
    cutoff_range : Tuple[float, float]
        Ranges of possible cutoffs in Hz
    q_range : Tuple[float, float]
        Range of possible Q factors

    Returns
    -------
    Tensor
        Filtered output of shape (..., T)
    """
    cutoff = _1rand_in_range(*cutoff_range)
    q = _1rand_in_range(*q_range)
    y = F.lowpass_biquad(x, sample_rate, cutoff, q)
    return y


def random_highpass(
    x: Tensor,
    sample_rate: int,
    cutoff_range: Tuple[float, float],
    q_range: Tuple[float, float],
) -> Tensor:
    """
    Apply a random highpass filter to the input batch.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (..., T)
    sample_rate : int
        Sample rate of the input
    cutoff_range : Tuple[float, float]
        Ranges of possible cutoffs in Hz
    q_range : Tuple[float, float]
        Range of possible Q factors

    Returns
    -------
    Tensor
        Filtered output of shape (..., T)
    """
    cutoff = _1rand_in_range(*cutoff_range)
    q = _1rand_in_range(*q_range)
    y = F.highpass_biquad(x, sample_rate, cutoff, q)
    return y


def random_peak_eq(
    x: Tensor,
    sample_rate: int,
    cutoff_range: Tuple[float, float],
    gain_range: Tuple[float, float],
    q_range: Tuple[float, float],
) -> Tensor:
    """
    Apply a random peak eq filter to the input batch.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (..., T)
    sample_rate : int
        Sample rate of the input
    cutoff_range : Tuple[float, float]
        Ranges of possible cutoffs in Hz
    q_range : Tuple[float, float]
        Range of possible Q factors

    Returns
    -------
    Tensor
        Filtered output of shape (..., T)
    """
    cutoff = _1rand_in_range(*cutoff_range)
    q = _1rand_in_range(*q_range)
    gain = _1rand_in_range(*gain_range)
    y = F.equalizer_biquad(x, sample_rate, cutoff, gain, q)
    return y



__all__ = [
    "load_audio",
    "save_audio",
    "load_audio_parallel",
    "load_audio_parallel_itr",
    "pack_audio_sequences",
]

from typing import Iterator, Sequence, Optional, Tuple
from resampy import resample as resample_np
from torchaudio.functional import resample
from multiprocess import Pool
from itertools import islice
from pathlib import Path
from torch import Tensor
import soundfile as sf
import numpy as np
import torchaudio
import torch


from torch_utilities.utilities import TensorOrArray


def load_audio(
    file_path: Path,
    sample_rate: int = None,
    tensor: bool = False,
    device: Optional[torch.device] = None,
) -> Tuple[TensorOrArray, int]:
    """
    Loads an audio file.

    Parameters
    ----------
    file_path : Path
        Path to the audio file
    sample_rate : int, optional
        Target sample rate, by default avoids resample
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : Optional[torch.device]
        The device to load the tensor into, by default None

    Returns
    -------
    Tuple[np.ndarray, int]
        (audio of shape (C, T), sample_rate)
    """
    # audio loading
    if not tensor:
        data, old_sample_rate = sf.read(
            file_path,
        )
        data = data.T
        if len(data.shape) == 1:
            data = data[None, :]
    else:
        data, old_sample_rate = torchaudio.load(file_path)
        if device is not None:
            data = data.to(device)

    # resampling
    if sample_rate is None:
        sample_rate = old_sample_rate
    elif old_sample_rate != sample_rate:
        if tensor:
            data = resample(data, old_sample_rate, sample_rate)
        else:
            data = resample_np(data, old_sample_rate, sample_rate)

    return data, sample_rate


def save_audio(file_path: Path, data: TensorOrArray, sample_rate: int):
    """
    Saves an audio file

    Parameters
    ----------
    file_path : Path
        Path to the audio file
    data : TensorOrArray
        Audio file to save
    sample_rate : int, optional
        Target sample rate
    """
    dtype = type(data)

    if dtype == np.ndarray:
        sf.write(file_path, data.T, samplerate=sample_rate)
    elif dtype == Tensor:
        data = data.to("cpu")
        torchaudio.save(file_path, data, sample_rate=sample_rate)
    else:
        err_msg = f"{dtype} is not supported by save_audio"
        raise NotImplementedError(err_msg)


def load_audio_parallel(
    file_paths: Sequence[Path],
    sample_rate: int = None,
    tensor: bool = False,
    device: Optional[torch.device] = None,
    num_workers: int = 4,
) -> Sequence[TensorOrArray]:
    """
    Loads multiple audio files in parallel.

    Parameters
    ----------
    file_path : Sequence[Path]
        Sequence of paths to the audio files
    sample_rate : int, optional
        Target sample rate, by default None
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : Optional[torch.device]
        The device to load the tensor into, by default None
    num_workers : int, optional
        Number of parallel processes

    Returns
    -------
    Sequence[TensorOrArray]
        Sequence of audios of shape (C, T)
    """
    with Pool(num_workers) as pool:
        _load = lambda x: load_audio(x, sample_rate, False)[0]
        xs = pool.map(_load, file_paths)
    if tensor:
        xs = [torch.tensor(x, device=device, dtype=torch.float32) for x in xs]
    return xs


def load_audio_parallel_itr(
    file_paths: Sequence[Path],
    sample_rate: int = None,
    tensor: bool = False,
    device: Optional[torch.device] = None,
    num_workers: int = 4,
) -> Iterator[TensorOrArray]:
    """
    Iterator that loads multiple audio files in parallel.

    Parameters
    ----------
    file_path : Sequence[Path]
        Sequence of paths to the audio files
    sample_rate : int, optional
        Target sample rate, by default None
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : Optional[torch.device]
        The device to load the tensor into, by default None
    num_workers : int, optional
        Number of parallel processes

    Returns
    -------
    Iterator[TensorOrArray]
        yields audios of shape (C, T)
    """
    n_files = len(file_paths)
    # transforming to generator
    file_paths = (x for x in file_paths)
    for i in range(0, n_files, num_workers):
        files_batch = islice(file_paths, num_workers)
        cache = load_audio_parallel(files_batch, sample_rate, tensor, device, num_workers)
        for x in cache:
            yield x


def pack_audio_sequences(
    xs: Sequence[Path],
    length: float,
    sample_rate: int,
    channels: int = 1,
    tensor: bool = False,
    delete_last: bool = True,
    num_workers: int = 1,
) -> Iterator[TensorOrArray]:
    """
    Reads from a Sequence of audio filepaths and generate temporal sequences
    of a certain length.

    Parameters
    ----------
    xs : Sequence[Path]
        Sequence of audio filepaths
    length : float
        Length of the sequences is seconds
    sample_rate : int
        Resample at this sample frequency
    channels : int, optional
        Number of channels of the sequences, by default 1
    tensor : bool, optional
        If False returns a numpy ndarray else a torch Tensor, by default False
    delete_last : bool, optional
        If True the last sequence is discarded (since it's typically not complete),
        by default True
    num_workers : int, optional
        Number of parallel processes

    Yields
    ------
    Iterator[TensorOrArray]
        Audio sequence of shape
        (length, C, T)
    """
    # length in samples
    length = int(length * sample_rate)

    # the container of the sequences to be generated
    _zeros = torch.zeros if tensor else np.zeros
    _reset_seq = lambda: _zeros((channels, length))

    # point to the last index consumed
    sample_ptr = 0
    seq_ptr = 0

    # utilities
    _seq_left = lambda: length - seq_ptr
    _sample_left = lambda x: x.shape[1] - sample_ptr
    _copy = lambda x: x.clone() if isinstance(x, Tensor) else x.copy()

    # enables multiprocessing ~ ~ ~ ~ ~
    if num_workers > 1:
        xs = load_audio_parallel_itr(xs, sample_rate, tensor, num_workers=num_workers)
    else:
        xs = (load_audio(x, sample_rate, tensor)[0] for x in xs)
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    seq = _reset_seq()
    for x in xs:
        while _sample_left(x) > 0:
            # copying into the sequence
            delta = min(_seq_left(), _sample_left(x))
            seq[:, seq_ptr : seq_ptr + delta] = x[:channels, sample_ptr : sample_ptr + delta]
            seq_ptr += delta
            sample_ptr += delta

            if _seq_left() == 0:
                # sequence complete
                yield _copy(seq)
                seq = _reset_seq()
                seq_ptr = 0

        # sample consumed
        sample_ptr = 0

    if not delete_last:
        yield _copy(seq)



import pytest
import torch

import torch_utilities.augmentation as aug
import torch_utilities as tu


# Local fixtures ===============================================================


@pytest.fixture(params=[-12, 0, 7])
def snr(request) -> int:
    """The signal-to-noise ratio."""
    return request.param


# scale
@pytest.fixture(params=[-3, 0])
def scale(request) -> int:
    """The scaling factor."""
    return request.param


# ==============================================================================


class TestAugmentation:
    def test_shuffle(self):
        x = torch.arange(1000)
        y = aug.shuffle(x)
        assert x.shape == y.shape
        assert torch.any(x.not_equal(y))

    def test_add_noise(self, snr):
        x = torch.ones((1, 1, 100))
        n = torch.ones_like(x)
        y, actual_snr = aug.add_noise(x, n, (snr, snr))
        assert x.max().equal(y.max())
        assert actual_snr - tu.invert_db(snr) < 1e-6
        assert len(actual_snr.shape) == 1
        assert actual_snr.shape[0] == y.shape[0]

    def test_scale(self, scale):
        batch_size = 4
        lin_scale = tu.invert_db(scale)
        x = torch.ones((batch_size, 1, 100))
        y, actual_scaling = aug.random_scaling(x, (scale, scale))
        assert y.shape == x.shape
        assert actual_scaling.shape == (x.shape[0],)
        torch.testing.assert_close(y.max().item(), lin_scale)
        torch.testing.assert_close(actual_scaling, torch.ones(batch_size) * lin_scale)

    def test_overdrive(self):
        x = torch.ones((4, 1, 100))
        y = aug.random_overdrive(x)
        torch.testing.assert_close(y.max().item(), 1.0)

    def test_dc_removal(self):
        x = torch.ones((2, 3, 5))
        y = aug.dc_removal(x)
        m = y.mean()
        assert m < 1e-6



__all__ = [
    "get_stft_istft",
    "MelFilterbank",
    "MelInverseFilterbank",
    "db",
    "invert_db",
    "power",
    "energy",
    "rms",
    "fade_sides",
    "random_trim",
    "trim_silence",
    "interleave",
    "trim_as_shortest",
]

from asteroid_filterbanks.transforms import from_torch_complex, to_torch_complex
from asteroid_filterbanks import STFTFB, Encoder, Decoder
from torchaudio.functional import melscale_fbanks
from typing import Callable, Optional, List, Tuple
from random import randrange
from torch import Tensor
import numpy as np
import torch


from torch_utilities.utilities import (
    get_np_or_torch,
    TensorOrArray,
    transpose,
    to_numpy,
)


def get_stft_istft(
    sample_rate: int,
    n_fft: int,
    hop_size: int,
    window: Optional[TensorOrArray] = None,
    complex: bool = True,
    pack_nyquist: bool = False,
) -> Tuple[Callable, Callable]:
    """Wrapper around asteroid filterbanks to get the STFT and ISTFT.

    Args:
        sample_rate (int): Sample rate of the signal
        n_fft (int): Number of FFT points
        hop_size (int): Hop size
        window (Optional[TensorOrArray], optional): Window to use. Defaults to Hann.
        complex (bool, optional): Whether to return complex numbers or dual-real. Defaults to True.
        pack_nyquist (bool, optional): Whether to pack the Nyquist frequency. Defaults to False.

    Note:
        - STFT in/out: (B, C, T) -> (B, C, T, 2*F)
        - ISTFT in/out: (B, C, T, 2*F) -> (B, C, T)
        - When complex==True the output is complex (F), otherwise dual-real (2*F).
        - When complex==False the real and imaginary parts are concatenated (not interleaved).
        - We support Tensor only

    Returns:
        Tuple[Callable, Callable]: stft, istft functions
    """
    if type(window) == np.ndarray:
        window = torch.from_numpy(window)
    if window is None:
        window = torch.hann_window(n_fft)

    kwargs = dict(
        n_filters=n_fft,
        kernel_size=n_fft,
        window=window,
        stride=hop_size,
        sample_rate=sample_rate,
    )
    stft = Encoder(STFTFB(**kwargs), as_conv1d=False)
    istft = Decoder(STFTFB(**kwargs))

    if complex:
        stft2 = lambda x: to_torch_complex(stft(x))
        istft2 = lambda x: istft(from_torch_complex(x))
    else:
        stft2 = stft
        istft2 = istft

    # We want (B, C, T, F)
    stft3 = lambda x: stft2(x).transpose(-1, -2)
    istft3 = lambda x: istft2(x.transpose(-1, -2))

    if pack_nyquist:
        raise NotImplementedError("Packing Nyquist is not implemented yet.")

    return stft3, istft3


# TODO: Basic tests
def pack_nyquist(x: TensorOrArray, complex: bool = True) -> TensorOrArray:
    """Move the real component of the Nyquist to the
    imaginary component of the 0 frequency.

    Args:
        x (TensorOrArray): Input signal of shape (..., F) or (..., 2*F)
        complex (bool, optional): Whether the input is complex or dual-real. Defaults to True.

    Returns:
        TensorOrArray: (..., F-1) or (..., 2*F-2)
    """
    if complex:
        x[..., 0] += 1j * x[..., -1].real
        x = x[..., :-1]
    else:
        freqs = x.shape[-1] // 2
        # niq.real -> 0.imag
        x[..., freqs] = x[..., freqs - 1]
        # removing niq.real and niq.imag
        x = x[..., :-1]
        x = torch.stack([x[..., : freqs - 1], x[..., freqs:]], -1)
    return x


def unpack_nyquist(x: TensorOrArray, complex: bool = True) -> TensorOrArray:
    pass  # TODO


class MelFilterbank:
    def __init__(
        self,
        sample_rate: int,
        n_freqs: int,
        n_mels: int,
        device: Optional[torch.device] = None,
    ):
        """
        Apply Mel filterbank to the input batch.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        n_freqs : int
            Stft frequencies
        n_mels : int
            Number of mel frequencies
        device : Optional[torch.device], optional
            Device for the filterbank matrix, by default None
        """
        self.sample_rate = sample_rate
        self.n_freqs = n_freqs
        self.n_mels = n_mels
        self.device = device
        self.filterbank = self._get_filterbank()

    def _get_filterbank(self) -> Tensor:
        """
        Gets the mel filterbank

        Returns
        -------
        Tensor
            Mel filterbank matrix
        """
        filterbank = melscale_fbanks(
            n_freqs=self.n_freqs,
            n_mels=self.n_mels,
            f_min=30,
            f_max=self.sample_rate // 2,
            sample_rate=self.sample_rate,
        )
        filterbank = filterbank.to(self.device)

        return filterbank

    def __call__(self, x: TensorOrArray) -> Tensor:
        """
        Parameters
        ----------
        x : TensorOrArray
            Signal of shape (B, C, T, n_freq)

        Returns
        -------
        Tensor
            STFT of shape (B, C, T, n_mel)
        """
        is_np = get_np_or_torch(x) == np
        if is_np:
            x = torch.from_numpy(x).to(self.device)

        # handling dtypes
        x = x.abs()
        self.filterbank = self.filterbank.to(x.dtype)

        y = x @ self.filterbank

        if is_np:
            y = to_numpy(y)
        return y


class MelInverseFilterbank:
    def __init__(
        self,
        sample_rate: int,
        n_freqs: int,
        n_mels: int,
        device: Optional[torch.device] = None,
    ):
        """
        Apply inverse Mel filterbank to the input batch,
        to get back a spectrogram.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        n_freqs : int
            Stft frequencies
        n_mels : int
            Number of mel frequencies
        device : Optional[torch.device], optional
            Device for the filterbank matrix, by default None
        """
        self.sample_rate = sample_rate
        self.n_freqs = n_freqs
        self.n_mels = n_mels
        self.device = device
        self.filterbank = self._get_filterbank()

    def _get_filterbank(self) -> Tensor:
        """
        Gets the mel filterbank

        Returns
        -------
        Tensor
            Mel filterbank matrix
        """
        filterbank = melscale_fbanks(
            n_freqs=self.n_freqs,
            n_mels=self.n_mels,
            f_min=30,
            f_max=self.sample_rate // 2,
            sample_rate=self.sample_rate,
        )
        # pseudo-inverse is used to approximate
        # the inverse transform
        filterbank = torch.linalg.pinv(filterbank)
        filterbank = filterbank.to(self.device)

        return filterbank

    def __call__(self, x: TensorOrArray) -> TensorOrArray:
        """
        Parameters
        ----------
        x : TensorOrArray
            Signal of shape (B, C, T, n_freq)

        Returns
        -------
        TensorOrArray
            STFT of shape (B, C, T, n_mel)
        """
        is_np = get_np_or_torch(x) == np
        if is_np:
            x = torch.from_numpy(x).to(self.device)

        # handling dtypes
        x = x.abs()
        self.filterbank = self.filterbank.to(x.dtype)

        y = x @ self.filterbank

        if is_np:
            y = to_numpy(y)
        return y


def db(x: TensorOrArray, eps: float = 1e-12) -> TensorOrArray:
    """
    Converts linear to dB

    Parameters
    ----------
    x : TensorOrArray
        Input signal amplitude
    eps : float
        Number summed to x before applying the logarithm,
        by default 1e-12

    Returns
    -------
    float
        Input in dB
    """
    module = get_np_or_torch(x)
    return 20 * module.log10(x + eps)


def invert_db(x: TensorOrArray, eps: float = 1e-12) -> TensorOrArray:
    """
    Converts dB to linear

    Parameters
    ----------
    x : TensorOrArray
        Input signal amplitude in dB

    Returns
    -------
    float
        Input inverting dB
    """
    return 10 ** (x / 20) - eps


def power(x: TensorOrArray) -> TensorOrArray | float:
    """
    Power of a signal, calculated for each channel.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        Power of the signal of shape (...)
    """
    module = get_np_or_torch(x)
    pwr = module.einsum("...t,...t->...", x, x.conj())
    return pwr


def energy(x: TensorOrArray) -> TensorOrArray | float:
    """
    Energy of a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        Energy of the signal of shape (...)
    """
    samples = x.shape[-1]
    return power(x) / samples


def rms(x: TensorOrArray) -> TensorOrArray | float:
    """
    RMS of a signal, calculated for each channel.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        RMS of the signal of shape (...)
    """
    e = energy(x)
    return np.sqrt(e)


def _win_to_sides(
    x: TensorOrArray,
    win: TensorOrArray,
    fade_len: int,
    direction: str,
) -> TensorOrArray:
    """
    Handler used to apply a window over the sides of
    a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., C, T)
    win : TensorOrArray
        Window
    fade_len : TensorOrArray
        Length of each fade in samples
    direction : str
        Indicates the sides in which the fade is applied,
        one between "left", "right" or "both"

    Returns
    -------
    TensorOrArray
        Faded output
    """
    # error handling
    err_msg = 'direction must be one between "left", "right" or "both"'
    assert direction in ["left", "right", "both"], err_msg

    if direction in ["left", "both"]:
        x[..., :fade_len] *= win[:fade_len]
    if direction in ["right", "both"]:
        x[..., -fade_len:] *= win[-fade_len:]
    return x


def fade_sides(
    x: TensorOrArray,
    fade_len: int = 100,
    direction: str = "both",
) -> TensorOrArray:
    """
    Apply an half of an Hanning window to the
    sides of the input, in order to obtain a fade in/out.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., C, T)
    fade_len : int, optional
        Length of the fade in samples, by default 10.
        The length of the window is 2 * fade_len + 1.
    direction : str
        Indicates the sides in which the fade is applied,
        one between "left", "right" or "both", by default "both"

    Returns
    -------
    TensorOrArray
        Faded output
    """
    module = get_np_or_torch(x)
    win = module.hanning(2 * fade_len + 1)
    if module == np:
        y = x.copy()
    else:
        win = win.to(x.device)
        win[-1] = 0
        y = x.detach().clone()
    y = _win_to_sides(y, win, fade_len, direction)

    return y


def random_trim(
    x: TensorOrArray,
    sample_rate: int,
    duration: float = 3,
) -> TensorOrArray:
    """
    Extracts a random temporal selection from the input.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., T)
    sample_rate : int
        Sample rate in Hz
    duration : float, optional
        Duration of the selection, by default 3 s

    Returns
    -------
    TensorOrArray
        Random temporal selection of the input
    """
    module = get_np_or_torch(x)
    x_len = x.shape[-1]
    duration_samples = int(duration * sample_rate)
    selection_start_max = x_len - duration_samples

    if selection_start_max <= 0:
        # if the input is longer than the selection
        # just zero pad the input
        y = module.zeros((*x.shape[:-1], duration_samples))
        y[..., :x_len] = x
    else:
        # applying selection
        start = randrange(0, selection_start_max)
        end = start + duration_samples
        y = x[..., start:end]
    return y


def trim_silence(
    x: TensorOrArray,
    threshold_db: float = -35,
    margin: int = 0,
) -> TensorOrArray:
    """
    Trims the silences at the beginning and end of a sample.

    Parameters
    ----------
    x : TensorOrArray
        Input sample of shape (T,)
    threshold_db : float, optional
        Relative to x.max() to detect the silences, by default -35 dB
    margin : int, optional
        Samples kept at both sides after the trimming, by default 0

    Returns
    -------
    TensorOrArray
        Trimmed ouput of shape (T',)
    """
    module = get_np_or_torch(x)

    # finding the start and end points
    threshold = invert_db(threshold_db)
    threshold *= module.abs(x).max()
    thr = module.zeros_like(x, dtype=int)
    thr[module.abs(x) > threshold] = 1
    thr = thr if module == np else thr.cpu().detach().numpy()
    thr = thr.tolist()

    try:
        start = thr.index(1)
        end = len(thr) - thr[::-1].index(1)

        # trimming the silences
        x_start = int(np.clip(start - margin, 0, None))
        x_end = end + margin
        y = x[..., x_start:x_end]
        return y
    except ValueError:
        # no value found
        return x


def interleave(*xs: List[TensorOrArray]) -> Tensor:
    """
    Interleaves many input tensors in one over the last dimension.

    Parameters
    ----------
    xs : List[TensorOrArray]
        Input signals of the same shape

    Returns
    -------
    Tensor
        Interleaved signal, the shape is the same of the inputs except
        for the interleaved dimension of lenght D that will be len(xs)*D
    """
    mod = get_np_or_torch(xs[0])
    if mod == np:
        stride = len(xs)
        new_shape = list(xs[0].shape)
        new_shape[-1] *= stride
        y = mod.zeros(new_shape)

        for i, x in enumerate(xs):
            y[..., i::stride] = x
    else:
        y = torch.stack(xs, -1).flatten(-2, -1)

    return y


def trim_as_shortest(*xs: List[TensorOrArray], dim: int = -1) -> List[TensorOrArray]:
    """
    Trims all the inputs to the same length of the shortest one.

    Parameters
    ----------
    xs : List[TensorOrArray]
        Input signals of the same shape

    dim : int, optional
        Dimension along which the trimming is applied, by default -1

    Returns
    -------
    List[TensorOrArray]
        Trimmed signals
    """
    min_len = min([x.shape[dim] for x in xs])
    xs = [transpose(x, dim, -1) for x in xs]
    xs = [x[..., :min_len] for x in xs]
    xs = [transpose(x, dim, -1) for x in xs]
    return xs



__all__ = [
    "OneOrPair",
    "TensorOrArray",
    "disable_cuda",
    "get_np_or_torch",
    "pack_many",
    "execute_with_probability",
    "to_numpy",
    "phase",
    "transpose",
    "factorize",
]

from typing import List, Sequence, Tuple, TypeVar, Union
from numpy import ndarray
from torch import Tensor
import numpy as np
import torch
import os


""" 
Generic variable
"""
T = TypeVar("T")

"""
Single or Pair of values of a certain type.
"""
OneOrPair = Union[T, Tuple[T, T]]

"""
Can be a torch Tensor or numpy ndarray.
"""
TensorOrArray = Union[Tensor, ndarray]


def disable_cuda():
    """Disables CUDA in PyTorch."""
    os.environ["CUDA_VISIBLE_DEVICES"] = ""


def get_np_or_torch(x: TensorOrArray):
    """
    Returns numpy or torch modules depending on the input

    Parameters
    ----------
    x : TensorOrArray
        Input

    Returns
    -------
    Module
        numpy or torch
    """
    if isinstance(x, Tensor):
        return torch
    else:
        torch.iscomplex = torch.is_complex
        torch.hanning = torch.hann_window
        np.cat = np.concatenate
        return np


def pack_many(*xss: List[List]) -> List[Tuple]:
    """
    Packs many Lists in one.

    Parameters
    ----------
    xss : List[List]

    Returns
    -------
    List[Tuple]
        Packed list
    """
    return list(zip(*xss))


def execute_with_probability(prob: float) -> bool:
    """
    Returns True of false based on the
    random variable prob

    Parameters
    ----------
    prob : float
        Probability to execute, from 0 (never happen)
        to 1 (always happen)

    Returns
    -------
    bool
        True prob % of the times
    """
    return np.random.uniform() > prob


def to_numpy(x: Tensor) -> np.ndarray:
    """
    Converts a Tensor into a numpy array.

    Parameters
    ----------
    x : Tensor
        Original tensor

    Returns
    -------
    np.ndarray
        Converted np array
    """
    return x.cpu().detach().numpy()


def phase(x: TensorOrArray) -> TensorOrArray:
    """
    Computes the phase of a complex signal x
    as in: phase(x) = exp{j*angle(x)}

    Parameters
    ----------
    x : TensorOrArray
        Input complex signal

    Returns
    -------
    TensorOrArray
        Elementwise phase
    """
    module = get_np_or_torch(x)
    y = module.exp(1j * module.angle(x))
    return y


def transpose(x: TensorOrArray, dim1: int, dim2: int) -> TensorOrArray:
    """
    Transpose a tensor over two dimensions.

    Parameters
    ----------
    x : TensorOrArray
        Input tensor
    dim1 : int, optional
        First dimension to transpose
    dim2 : int, optional
        Second dimension to transpose

    Returns
    -------
    TensorOrArray
        Transposed tensor
    """
    return x.transpose(dim1, dim2) if type(x) == Tensor else np.swapaxes(x, dim1, dim2)


def factorize(n: int) -> Sequence[int]:
    """
    Factorize an integer number.
    Implementation based on
    https://stackoverflow.com/questions/16007204/factorizing-a-number-in-python

    Parameters
    ----------
    n : int
        Number to factorize

    Returns
    -------
    Sequence[int]
        List of factors in increasing order
    """

    def get_factor(n):
        j = 2
        while n > 1:
            for i in range(j, int(np.sqrt(n + 0.05)) + 1):
                if n % i == 0:
                    n //= i
                    j = i
                    yield i
                    break
            else:
                if n > 1:
                    yield int(n)
                    break

    factors = list(get_factor(n))
    factors = sorted(factors)
    return factors



from typing import Callable, Tuple
import numpy as np
import pytest
import torch

from torch_utilities import TensorOrArray
import torch_utilities as tu

# Local fixtures ===============================================================


@pytest.fixture(params=[160])
def n_freq(request) -> int:
    """The number of frequencies."""
    return request.param


@pytest.fixture(params=[32])
def n_mel(request) -> int:
    """The number of mel filters."""
    return request.param


@pytest.fixture
def mel_filterbank(sample_rate, n_freq, n_mel) -> tu.MelFilterbank:
    """Mel filterbank instance."""
    return tu.MelFilterbank(sample_rate, n_freq, n_mel)


@pytest.fixture
def mel_inverse_filterbank(sample_rate, n_freq, n_mel) -> tu.MelInverseFilterbank:
    """Mel filterbank instance."""
    return tu.MelInverseFilterbank(sample_rate, n_freq, n_mel)


def _input_common(module, channels, sample_rate, f) -> TensorOrArray:
    x = np.ones((1, channels, int(sample_rate * 0.1), f), dtype=complex)
    if module == torch:
        x = torch.from_numpy(x)
    return x


@pytest.fixture
def input_spec(module, channels, sample_rate, n_freq) -> TensorOrArray:
    """Input spectrogram."""
    return _input_common(module, channels, sample_rate, n_freq)


@pytest.fixture
def input_mel(module, channels, sample_rate, n_mel) -> TensorOrArray:
    """Input mel."""
    return _input_common(module, channels, sample_rate, n_mel)


@pytest.fixture
def rand_float(module) -> TensorOrArray:
    x = np.random.uniform(0, 1, 1)
    if module == torch:
        x = torch.tensor(x)
    return x


@pytest.fixture(params=[0, 10])
def trim_margin(request) -> int:
    """Trim margin."""
    return request.param


@pytest.fixture(params=[128])
def n_fft(request) -> int:
    """Number of FFT points."""
    return request.param


@pytest.fixture
def n_stft_freqs(n_fft) -> int:
    """Number of FFT frequencies."""
    return n_fft // 2 + 1


@pytest.fixture()
def hop_size(n_fft) -> int:
    """Hop size."""
    return n_fft // 2


@pytest.fixture(params=[True, False])
def use_complex(request) -> bool:
    """Use complex."""
    return request.param


@pytest.fixture
def stft_istft(sample_rate, n_fft, hop_size, use_complex) -> Tuple[Callable, Callable]:
    """STFT and ISTFT functions."""
    stft, istft = tu.get_stft_istft(
        sample_rate=sample_rate,
        n_fft=n_fft,
        hop_size=hop_size,
        complex=use_complex,
    )

    return stft, istft


@pytest.fixture
def stft(stft_istft) -> Callable:
    """STFT function."""
    stft, _ = stft_istft
    return stft


@pytest.fixture
def istft(stft_istft) -> Callable:
    """ISTFT function."""
    _, istft = stft_istft
    return istft


# ==============================================================================


class TestSTFT:
    def test_stft_shape(self, stft, use_complex, n_stft_freqs):
        x = torch.zeros(1, 1, 16000)
        y = stft(x)
        assert y.shape[:2] == x.shape[:2]
        assert len(y.shape) == 4
        if use_complex:
            assert torch.is_complex(y)
            assert y.shape[-1] == n_stft_freqs
        else:
            assert not torch.is_complex(y)
            assert y.shape[-1] == 2 * n_stft_freqs

    def test_istft_shape(self, istft, use_complex, n_stft_freqs):
        x = torch.zeros(1, 1, 10, 2 * n_stft_freqs)
        if use_complex:
            x = x[..., :n_stft_freqs] + 1j
        y = istft(x)
        assert not torch.is_complex(y)
        assert y.shape[:2] == (1, 1)
        assert len(y.shape) == 3


class TestMelFilterbank:
    def test_to_mel_shape(self, mel_filterbank, input_spec, n_mel):
        x = input_spec
        y = mel_filterbank(x)
        assert x.shape[:-1] == y.shape[:-1]
        assert y.shape[-1] == n_mel

    def test_to_freq_shape(self, mel_inverse_filterbank, input_mel, n_freq):
        x = input_mel
        y = mel_inverse_filterbank(x)
        assert x.shape[:-1] == y.shape[:-1]
        assert y.shape[-1] == n_freq


class TestAudio:
    def test_db(self, rand_float):
        x = rand_float
        y = tu.db(x)
        x_hat = tu.invert_db(y)
        torch.testing.assert_close(x, x_hat)

    def test_power(self, module):
        x = module.ones((2, 100))
        x[50:] = -1
        torch.testing.assert_close(tu.power(x), module.ones(2) * 100.0)

    def test_energy(self, module):
        x = module.ones((2, 100))
        x[50:] = -1
        torch.testing.assert_close(tu.energy(x), module.ones(2) * 1.0)

    def test_rms(self, module):
        x = module.ones((2, 100))
        x[50:] = -1
        torch.testing.assert_close(tu.rms(x), module.ones(2) * 1.0)

    def test_fade_sides_2d(self, module):
        x = module.ones((2, 200))
        x = tu.fade_sides(x)
        for i in range(2):
            zero = module.zeros(1)
            torch.testing.assert_close(x[i, 0], zero[0])
            torch.testing.assert_close(x[i, -1], zero[0])

    def test_trim(self, module):
        x = module.zeros((1, 1, 5 * 160))
        y = tu.random_trim(x, 160, 2)
        assert y.shape == (1, 1, 2 * 160)

    def test_trim_silence(self, module, trim_margin):
        x = module.zeros(100)
        x[30:40] = 0.1
        y = tu.trim_silence(x, margin=trim_margin)
        assert y.shape[-1] == (10 + 2 * trim_margin)

    def test_interleave(self, module):
        dimensions = 3
        freqs = 10
        x_size = [freqs if i == (dimensions - 1) else 1 for i in range(dimensions)]

        if module == np:
            x = np.random.normal(size=x_size)
        else:
            x = torch.randn(x_size)

        xs = [x for _ in range(dimensions)]
        y = tu.interleave(*xs)
        y_dim = x_size
        y_dim[-1] *= dimensions
        assert y.shape == tuple(y_dim)
        for i, z in enumerate(xs):
            e = module.abs(y[..., i::dimensions] - z).max()
            assert e < 1e-6

    def test_trim_as_shortest(self, module):
        x = module.ones((1, 100, 1))
        y = module.ones((1, 50, 1))
        x, y = tu.trim_as_shortest(x, y, dim=1)
        assert x.shape == (1, 50, 1)
        assert y.shape == (1, 50, 1)



__all__ = ["DNSMOS"]

# FIXME

import onnxruntime as ort
from typing import Tuple
from pathlib import Path
from torch import Tensor
import numpy as np


from torch_utilities.utilities import to_numpy, TensorOrArray


def _get_model_dir() -> Path:
    """
    Returns the directory containing the models.

    Returns
    -------
    Path
        Path to the directory
    """
    path = Path(__file__).parent / "models"
    return path


class DNSMOS:
    def __init__(self, onnx_exec_provider: str = "CPUExecutionProvider"):
        """Microsoft DNSMOS model.

        Args:
            onnx_exec_provider (str, optional): Exectution provider used for ONNXRuntime. Defaults to "CPUExecutionProvider".
        """
        self.model_path = _get_model_dir() / "sig_bak_ovr.onnx"
        self.onnx_exec_provider = onnx_exec_provider
        self.sample_rate = 16000  # Hz
        self.hopsize = 1 * self.sample_rate
        self.in_len = int(9.01 * self.sample_rate)
        self.inference_sess = self.get_inference_session()

    def get_inference_session(self) -> ort.InferenceSession:
        """
        Returns an inference session for the model.

        Returns
        -------
        ort.InferenceSession
            DNSMOS inference session
        """
        session = ort.InferenceSession(
            str(self.model_path),
            providers=[self.onnx_exec_provider],
        )
        return session

    def get_polyfit_val(
        self, ovr: float, sig: float, bak: float
    ) -> Tuple[float, float, float]:
        """
        Polynomial nonlinearity as in the original repo

        Parameters
        ----------
        ovr : float
            Ovr raw value
        sig : float
            Sig raw value
        bak : float
            Bak raw value

        Returns
        -------
        Tuple[float, float, float]
            Processed metrics
        """
        p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
        p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
        p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        ovr_poly = p_ovr(ovr)
        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)

        return ovr_poly, sig_poly, bak_poly

    def __call__(self, x: TensorOrArray) -> Tuple[float, float, float]:
        """
        Compute ovr, sig and bak from an 16kHz input.

        Parameters
        ----------
        x : TensorOrArray
            Input of shape (C, T) sampled at 16kHz,
            if multichannel, the first channel only is considered

        Returns
        -------
        Tuple[float, float, float]
            (ovr, sig, bak)
        """
        assert len(x.shape) == 2, "the shape of x should be (C, T)"

        if isinstance(x, Tensor):
            x = to_numpy(x)
        x = x.astype("float32")

        # keeping the first channel
        x = x[:1]

        # computing the average over segments of in_len shifting by hopsize
        n_frames = (
            np.clip(x.shape[1] - self.in_len, 0, None).astype(int) // self.hopsize + 1
        )
        ovr = []
        sig = []
        bak = []
        for i in range(n_frames):
            a = int(i * self.sample_rate)
            b = int(a + self.in_len)
            x_sel = x[:, a:b]
            # Padding if needed
            if x_sel.shape[1] != self.in_len:
                x_sel = np.pad(x_sel, ((0, 0), (0, self.in_len - x_sel.shape[1])))
            model_in = dict(input_1=x_sel)
            scores = self.inference_sess.run(None, model_in)[0][0]
            scores = self.get_polyfit_val(*scores)
            [container.append(s) for container, s in zip((sig, bak, ovr), scores)]
        ovr, sig, bak = [np.mean(container) for container in (ovr, sig, bak)]

        return ovr, sig, bak



from torch import Tensor
import numpy as np
import pytest
import torch

import torch_utilities as tu
from torch_utilities import TensorOrArray

# Local fixtures ===============================================================


@pytest.fixture(params=[True, False])
def remove_last(request) -> bool:
    """Remove last."""
    return request.param


# ==============================================================================


class TestIO:
    def test_load_audio(self, audio_file, channels, sample_rate, module):
        is_tensor = module == torch
        expected_type = Tensor if is_tensor else np.ndarray
        x, x_sr = tu.load_audio(audio_file, sample_rate, is_tensor)
        assert len(x.shape) == 2
        assert x.shape[0] == channels
        assert type(x) == expected_type
        assert x_sr == sample_rate

    def test_save_audio(self, channels, sample_rate, module, session_tmp_dir):
        out_path = session_tmp_dir / "save_audio_out.wav"
        out_path.unlink(missing_ok=True)
        x = module.zeros((channels, sample_rate // 16))
        tu.save_audio(out_path, x, sample_rate)
        assert out_path.exists()

    def test_load_audio_parallel(
        self, audio_file, channels, sample_rate, module, workers
    ):
        n_files = 2 * workers
        is_tensor = module == torch
        expected_type = Tensor if is_tensor else np.ndarray
        filepaths = [audio_file for _ in range(n_files)]
        xs = tu.load_audio_parallel(
            filepaths, sample_rate, is_tensor, num_workers=workers
        )
        assert len(xs) == n_files
        for x in xs:
            assert type(x) == expected_type
            assert x.shape[0] == channels

    def test_load_audio_parallel_itr(
        self, audio_file, channels, sample_rate, module, workers
    ):
        n_files = 2 * workers
        is_tensor = module == torch
        expected_type = Tensor if is_tensor else np.ndarray
        filepaths = [audio_file for _ in range(n_files)]
        xs = tu.load_audio_parallel_itr(
            filepaths, sample_rate, is_tensor, num_workers=workers
        )
        for i, x in enumerate(xs):
            assert type(x) == expected_type
            assert x.shape[0] == channels
        assert i + 1 == n_files

    def test_pack_audio_sequences(
        self, audio_file, workers, channels, sample_rate, module, remove_last
    ):
        length = 1.5
        tensor = module == torch
        n_files = 10

        files = [audio_file for _ in range(n_files)]
        itr = tu.pack_audio_sequences(
            xs=files,
            length=length,
            sample_rate=sample_rate,
            channels=channels,
            tensor=tensor,
            delete_last=remove_last,
            num_workers=workers,
        )
        for i, seq in enumerate(itr):
            assert seq.shape == (channels, int(sample_rate * length))
        n = i + 1 if remove_last else i
        assert n == int(n_files / length)



__all__ = [
    # utilities
    "LambdaLayer",
    "LookAhead",
    "Reparameterize",
    "ScaleChannels2d",
    "UnfoldSpectrogram",
    "FoldSpectrogram",
    "ResidualWrap",
    # dense variants
    "GroupedLinear",
    # conv2d variants
    "CausalConv2d",
    "CausalSubConv2d",
    # conv2d compositions
    "CausalConv2dNormAct",
    "DenseConvBlock",
    # recurrent variants
    "GruNormAct",
    # attention variants
    "SlidingCausalMultiheadAttention",
]

from typing import Callable, List, Optional, Tuple
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn import Module
import numpy as np
import torch


from torch_utilities.audio import interleave
from torch_utilities.utilities import OneOrPair


# private utility functions = = = = = = = = = = = = = =


def get_time_value(param):
    """
    Extracts the parameter referring to the
    temporal axis.

    Parameters
    ----------
    param : tuple or scalar
        Module parameter

    Returns
    -------
    scalar
        Temporal parameter
    """
    if type(param) in (tuple, list):
        return param[0]
    else:
        return param


def get_freq_value(param):
    """
    Extracts the parameter referring to the
    frequency axis.

    Parameters
    ----------
    param : tuple or scalar
        Module parameter

    Returns
    -------
    scalar
        Temporal parameter
    """
    if type(param) in (tuple, list):
        return param[1]
    else:
        return param


def get_causal_conv_padding(kernel_size: int, dilation: int) -> int:
    """
    Calculates the causal convolutional padding.

    Returns
    -------
    int
       Total causal padding
    """
    causal_pad = (kernel_size - 1) * dilation
    return causal_pad


def get_default_dilation(
    kernel_size: OneOrPair[int],
    depth: int,
    disable_dilation_f: bool = False,
) -> List[Tuple[int, int]]:
    """
    The default dilation is an increasing power of the kernel
    # e.g: d_t_conv2 = k_t**2

    Parameters
    ----------
    kernel_size : OneOrPair[int]
        Size of the kernel
    depth : int
        Number of layers with dilation
    disable_dilation_f : bool, optional
        If True the dilation over frequency will be always 1,
        by default False

    Returns
    -------
    List[Tuple[int, int]]
        default dilation
    """
    k_t, k_f = [f(kernel_size) for f in (get_time_value, get_freq_value)]
    dilation = [(k_t**x, k_f**x) for x in range(depth)]
    if disable_dilation_f:
        dilation = [(d[0], 1) for d in dilation]
    return dilation


def get_causal_longformer_mask(sequence_len: int, receptive_field: int) -> Tensor:
    """
    Obtain a causal mask inspired by the Longformer to use
    within self attention layers.

    https://arxiv.org/abs/2004.05150

    Parameters
    ----------
    sequence_len : int
        Length of the sequence
    receptive_field : int
        Number of previous frames to pay attention to

    Returns
    -------
    Tensor
        Attention mask of shape (sequence_len, sequence_len)
    """
    _diag = lambda i: torch.diag(torch.ones(sequence_len - i), -i)
    mask = _diag(0)
    for i in range(1, receptive_field):
        mask += _diag(i)
    return mask


# utilitiy layers  = = = = = = = = = = = = = = = = = = =
class LambdaLayer(Module):
    def __init__(self, f: Callable):
        """
        Inspired to TF lambda layer

        Parameters
        ----------
        f : Callable
            Callable function
        """
        super(LambdaLayer, self).__init__()
        self.f = f

    def forward(self, *x):
        return self.f(*x)


class LookAhead(Module):
    def __init__(
        self,
        lookahead: int,
        maintain_shape: bool = False,
    ):
        """
        Temporal lookahead layer.
        Input shape: (B, C, T, F)
        Output shape: (B, C, T', F)

        Parameters
        ----------
        lookahead : int
            Amount of lookahead
        maintain_shape : bool, optional
            If set to True, right zero padding is add to compensate
            for the lookahead, by default False
        """
        super().__init__()

        # attributes
        self.lookahead = lookahead
        self.maintain_shape = maintain_shape

        # inner modules
        if self.maintain_shape:
            self.lookahead_pad = nn.ConstantPad2d(
                (0, 0, -self.lookahead, self.lookahead), 0
            )
        else:
            self.lookahead_pad = nn.ConstantPad2d((0, 0, -self.lookahead, 0), 0)

    def forward(self, x: Tensor) -> Tensor:
        x = self.lookahead_pad(x)
        return x


class Reparameterize(Module):
    def __init__(self):
        super().__init__()

    def forward(self, mu: Tensor, logvar: Tensor) -> Tensor:
        """
        Reparameterization trick to sample
        from N(mu, var) without breaking the gradient path.

        Parameters
        ----------
        mu : Tensor
            Mean of the latent normal
        logvar : Tensor
            Standard deviation of the latent normal in log scale

        Returns
        -------
        Tensor
            Sample from the normal distribution
        """
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps * std + mu


class ScaleChannels2d(Module):
    def __init__(
        self,
        in_channels: int,
    ):
        """
        Learns a per-channel gain.

        Parameters
        ----------
        in_channels : int
            Number of channels
        """
        super().__init__()

        # attributes
        self.in_channels = in_channels

        # inner modules
        self.scale = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.in_channels,
            kernel_size=1,
            groups=self.in_channels,
            bias=False,
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Applies a gain to the channels of the input.

        Parameters
        ----------
        x : Tensor
            Input of shape (B, C, T, F)

        Returns
        -------
        Tensor
            Scaled input of shape (B, C, T, F)
        """
        x = self.scale(x)
        return x


class UnfoldSpectrogram(Module):
    def __init__(self, block_size: int, stride: int):
        """
        Custom unfold module, it performs a windowing over time of a time-freq signal.

        Parameters
        ----------
        block_size : int
            Size of the sliding window over time
        stride : int
            Stride of the sliding window over time
        """
        super().__init__()

        # attributes
        self.block_size = block_size
        self.stride = stride

        # inner modules ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

        # (B, C, T, F) -> (B*C, 1, T, F)
        self._reshape_0 = lambda x: x.reshape(x.shape[0] * x.shape[1], *x.shape[2:])[
            :, None, :, :
        ]

        # (B*C, 1, T, F) -> (B*C, F*block_size, num_blocks)
        self._unfold = lambda x: F.unfold(
            x, (self.block_size, x.shape[3]), stride=(self.stride, 1)
        )

        # (B*C, block_size*F, num_blocks) -> (B*C, 1, block_size, F, num_blocks)
        self._reshape_1 = lambda x: x.reshape(
            x.shape[0], 1, self.block_size, -1, x.shape[2]
        )

        # (B*C, 1, block_size, F, num_blocks) -> (B*C, 1, num_blocks, block_size, F)
        self._reshape_2 = lambda x: x.permute(0, 1, 4, 2, 3)

        # (B*C, 1, num_blocks, block_size, F) -> (B, C, num_blocks, block_size, F)
        self._reshape_3 = lambda x, ch: x.reshape(
            -1, ch, x.shape[2], self.block_size, x.shape[4]
        )

        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Input of shape (B, C, T, F)

        Returns
        -------
        Tensor
            Unfolded input of shape (B, C, num_blocks, block_size, F)
        """
        assert self._input_shape_is_valid(x), r"(x.shape[2] - block_size) % stride != 0"
        ch = x.shape[1]
        x = self._reshape_0(x)
        x = self._unfold(x)
        x = self._reshape_1(x)
        x = self._reshape_2(x)
        x = self._reshape_3(x, ch)
        return x

    def _input_shape_is_valid(self, x: Tensor) -> bool:
        """
        Verifies that the input shape is valid for unfolding

        Parameters
        ----------
        x : Tensor
            Input tensor

        Returns
        -------
        bool
            True if the input shape is valid
        """
        t = x.shape[2]
        flag = (t - self.block_size) % self.stride == 0
        return flag


class FoldSpectrogram(Module):
    def __init__(self, block_size: int, stride: int, channels: int):
        """
        Custom fold module, reverses the result of UnfoldSpectrogram.

        Parameters
        ----------
        block_size : int
            Size of the sliding window over time
        stride : int
            Stride of the sliding window over time
        channels : int
            Number of channels before UnfoldSpectrogram
        """
        super().__init__()

        # attributes
        self.block_size = block_size
        self.stride = stride
        self.channels = channels

        # inner modules

        # (B, C*num_blocks, block_size, F)
        # used to get the output_size for the folding
        self._get_out_size = lambda x: (
            x.shape[1] * self.stride // self.channels + self.block_size - self.stride,
            x.shape[3],
        )

        # inner modules ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

        # (B, C, num_blocks, block_size, F) -> (B, C*num_blocks, block_size, F)
        self._reshape_0 = lambda x: x.reshape(x.shape[0], -1, *x.shape[3:])

        # (B, C*num_blocks, block_size, F) -> (B*C, 1, num_blocks, block_size, F)
        self._reshape_1 = lambda x: x.reshape(
            -1, 1, x.shape[1] // self.channels, self.block_size, x.shape[3]
        )

        # (B*C, 1, num_blocks, block_size, F) -> (B*C, 1, block_size, F, num_blocks)
        self._reshape_2 = lambda x: x.permute(0, 1, 3, 4, 2)

        # (B*C, 1, block_size, F, num_blocks) -> (B*C, block_size*F, num_blocks)
        self._reshape_3 = lambda x: x.reshape(x.shape[0], -1, x.shape[4])

        # (B*C, block_size*F, num_blocks) -> (B*C, 1, T, F)
        self._fold = lambda x, o: F.fold(
            x, o, (self.block_size, o[1]), stride=(self.stride, 1)
        )

        # (B*C, 1, T, F) -> (B, C, T, F)
        self._reshape_4 = lambda x: x.reshape(-1, self.channels, *x.shape[2:])

        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Input of shape (B, C, num_blocks, block_size, F)

        Returns
        -------
        Tensor
            Folded input of shape (B, C, T, F)
        """
        x = self._reshape_0(x)
        out_sizes = self._get_out_size(x)
        x = self._reshape_1(x)
        x = self._reshape_2(x)
        x = self._reshape_3(x)
        x = self._fold(x, out_sizes)
        x = self._reshape_4(x)
        divisor = self._get_normalization_tensor(x, out_sizes)
        x /= divisor
        return x

    def _get_normalization_tensor(
        self, x: Tensor, out_sizes: Tuple[int, int]
    ) -> Tensor:
        """
        Obtain the normalization tensor for perfect fold/unfold inversion.
        https://pytorch.org/docs/stable/generated/torch.nn.Unfold.html?highlight=unfold#torch.nn.Unfold

        Parameters
        ----------
        x : Tensor
            Input tensor
        out_sizes : Tuple[int, int]
            Parameter output_size needed for folding


        Returns
        -------
        Tensor
            Normalization tensor
        """
        o = out_sizes
        y = torch.ones_like(x)
        y = F.unfold(y, (self.block_size, o[1]), stride=(self.stride, 1))
        y = F.fold(y, o, (self.block_size, o[1]), stride=(self.stride, 1))
        return y


class ResidualWrap(Module):
    def __init__(self, *modules: List[Module]):
        """
        Given a series of layers F(x) it reparametrize
        them as (x + F(x)).

        F(x) and x shoud have compatible shapes.

        Parameters
        ----------
        modules : List[Module]
            Series of modules
        """
        super().__init__()
        self.layers = nn.Sequential(*modules)

    def forward(self, x: Tensor) -> Tensor:
        y = self.layers(x)
        # keeping the channels of the output
        c_out = y.shape[1]
        y = x[:, :c_out] + y[:, :c_out]

        return y


# dense variants = = = = = = = = = = = = = = = = = = = =
class GroupedLinear(Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        groups: int = 1,
    ):
        """
        Dense layer with grouping.

        Parameters
        ----------
        input_dim : int
            Input feature size
        output_dim : int
            Output feature size
        groups : int, optional
            number of groups, by default 1
        """
        super().__init__()

        # attributes
        self.input_dim = input_dim
        self.hidden_size = output_dim
        self.ws = input_dim // groups
        self.groups = groups

        # error handling
        assert (
            input_dim % groups == 0
        ), f"Input size {input_dim} not divisible by {groups}"
        assert (
            output_dim % groups == 0
        ), f"Hidden size {output_dim} not divisible by {groups}"

        # weights
        self.register_parameter(
            "weight",
            nn.Parameter(
                torch.zeros(groups, input_dim // groups, output_dim // groups),
                requires_grad=True,
            ),
        )
        self.reset_parameters()

    def reset_parameters(self):
        """
        Resets the dense weights.
        """
        nn.init.kaiming_uniform_(self.weight, a=np.sqrt(5))

    def forward(self, x: Tensor) -> Tensor:
        """
        Applies self.groups different dense layers
        distributed per sections.

        Parameters
        ----------
        x : Tensor
            Input of shape (..., self.input_dim)

        Returns
        -------
        Tensor
            Output of shape (..., self.hidden_size)
        """
        new_shape = (*x.shape[:-1], self.groups, self.ws)
        x = x.reshape(new_shape)
        x = torch.einsum("...gi,gih->...gh", x, self.weight)  # [..., G, H/G]
        x = x.flatten(-2, -1)
        return x

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}(input_dim: {self.input_dim}, hidden_size: {self.hidden_size}, groups: {self.groups})"


# conv2d variants = = = = = = = = = = = = = = = = = = =
class CausalConv2d(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: OneOrPair[int],
        stride_f: int = 1,
        padding_f: Optional[int] = None,
        dilation: OneOrPair[int] = 1,
        bias: bool = True,
        separable: bool = False,
        dtype=None,
    ):
        """
        Convolution with causal kernels over time

        Parameters
        ----------
        Same parameters as Conv2d plus

        stride_f : int
            Kernel stride over frequency
        padding_f : Optional[int]
            Symmetric padding over frequency, by default ensures f_out = f_in // stride_f
            if stride_f divides f_in
        separable : bool, optional
            Enable separable convolution (depthwise + pointwise), by default False
        """
        super().__init__()

        # attributes
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride_f = stride_f
        self.padding_f = padding_f
        self.dilation = dilation
        self.bias = bias
        self.separable = separable
        self.dtype = dtype

        # inner modules
        self.causal_pad_amount = self._get_causal_pad_amount()
        self.groups = np.gcd(self.in_channels, self.out_channels)

        layers = []

        # causal padding
        causal_pad = nn.ConstantPad2d((0, 0, self.causal_pad_amount, 0), 0)
        layers.append(causal_pad)

        # freq padding
        freq_pad = self._get_default_freq_padding()
        layers.append(freq_pad)

        # convolution
        conv_pad = 0 if self.padding_f is None else (0, self.padding_f)
        if not self.separable:
            full_conv = nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                kernel_size=self.kernel_size,
                stride=(1, stride_f),
                dilation=self.dilation,
                bias=self.bias,
                dtype=self.dtype,
                padding=conv_pad,
            )

            layers.append(full_conv)
        else:
            # separable convolution
            # depthwise + pointwise
            depthwise = nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.in_channels,
                kernel_size=self.kernel_size,
                stride=(1, stride_f),
                dilation=self.dilation,
                bias=self.bias,
                groups=self.groups,
                dtype=self.dtype,
                padding=conv_pad,
            )

            pointwise = nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                kernel_size=1,
                bias=False,
                dtype=self.dtype,
            )

            layers += [depthwise, pointwise]

        self.layers = nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.layers(x)
        return x

    def _get_causal_pad_amount(self) -> int:
        """
        Calculates the causal padding.
        """
        kernel_size_t, dilation_t = map(
            get_time_value, (self.kernel_size, self.dilation)
        )
        causal_pad = get_causal_conv_padding(kernel_size_t, dilation_t)
        return causal_pad

    def _get_default_freq_padding(self) -> Module:
        """
        Gets the default frequency padding.

        Returns
        -------
        Module
            Frequency padding module
        """
        kernel_size_f, dilation_f = map(
            get_freq_value, (self.kernel_size, self.dilation)
        )
        pad_f = dilation_f * (kernel_size_f - 1) + 1 - self.stride_f
        half_pad_f = pad_f // 2
        if self.padding_f is not None:
            pad = nn.Identity()  # manual padding
        elif pad_f % 2 == 0:
            pad = nn.ConstantPad2d((half_pad_f, half_pad_f, 0, 0), 0)
        else:
            pad = nn.ConstantPad2d((half_pad_f + 1, half_pad_f, 0, 0), 0)
        return pad


class CausalSubConv2d(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: OneOrPair[int],
        stride: int = 1,
        dilation: OneOrPair[int] = 1,
        bias: bool = True,
        separable: bool = False,
        dtype=None,
    ):
        """
        Also known as sub pixel convolution, described in
        Dense CNN With Self-Attention for Time-Domain Speech Enhancement
        paper (https://ieeexplore.ieee.org/document/9372863).

        Many convolutions are interleaved to upsample a signal over frequency or time.

        Parameters
        ----------
        Same parameters as Conv2d plus

        stride : int
            Defines how many interleaved convolutions are present
        separable : bool, optional
            Enable separable convolution (depthwise + pointwise), by default False
        """
        super().__init__()

        # attributes
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.bias = bias
        self.separable = separable
        self.dtype = dtype

        # inner modules
        _conv = lambda: CausalConv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride_f=1,
            padding_f=None,
            dilation=dilation,
            bias=bias,
            separable=separable,
            dtype=dtype,
        )
        self.layers = nn.Sequential(*[_conv() for _ in range(self.stride)])
        self.transpose = lambda x: x.transpose(2, 3)

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Input of shape (B, C, T, F)

        Returns
        -------
        Tensor
            Output of shape (B, C, T, F * stride)
        """
        xs = [conv(x) for conv in self.layers]
        x = interleave(*xs)
        return x


# conv2d compositions = = = = = = = = = = = = = = = = =
class CausalConv2dNormAct(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: OneOrPair[int],
        stride_f: int = 1,
        dilation: OneOrPair[int] = 1,
        separable: bool = False,
        batchnorm_eps: float = 1e-05,
        batchnorm_momentum: float = 0.1,
        batchnorm_affine: bool = True,
        batchnorm_track_running_stats: bool = True,
        activation: Module = nn.ReLU(),
        residual_merge: Optional[Callable] = None,
        merge_after_conv: bool = True,
        disable_batchnorm: bool = False,
        dtype=None,
    ):
        """
        CausalConv2d + BatchNorm2d + Activation.

        This layer ensures f_in = f_out // stride_f only if stride_f divides f_in.

        Parameters
        ----------
        in_channels : int
            Number of input channels
        out_channels : int
            Number of output channels
        kernel_size : OneOrPair[int]
            Convolution kernel size
        stride_f : int, optional
            Frequency stride, by default 1
        dilation : OneOrPair[int], optional
            Convolution dilation, by default 1
        separable : bool, optional
            Enable separable convolution (depthwise + pointwise), by default False
        batchnorm_eps : float, optional
            Eps parameter of the BatchNorm2d , by default 1e-05
        batchnorm_momentum : float, optional
            Momentum parameter of the BatchNorm2d, by default 0.1
        batchnorm_affine : bool, optional
            Affine parameter of the BatchNorm2d, by default True
        batchnorm_track_running_stats : bool, optional
            Track running stats parameter of the BatchNorm2d, by default True
        activation : Module, optional
            Activation module, by default nn.Relu()
        residual_merge : Optional[Callable], optional
            If different da None, it indicates the merge operation using a skip connection
            from the output of the transposed conv to the output of the activation, by default None
        merge_after_conv : bool, optional
            If True one the input used in the residual_merge layer is taken after the CausalConv2d,
            instead that from the input directly, by default True
        disable_batchnorm : bool, optional
            Disable the BatchNorm2d layer, by default False
        dtype, optional
            Module dtype, by default None
        """

        super().__init__()

        # attributes
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride_f = stride_f
        self.dilation = dilation
        self.separable = separable
        self.batchnorm_eps = batchnorm_eps
        self.batchnorm_momentum = batchnorm_momentum
        self.batchnorm_affine = batchnorm_affine
        self.batchnorm_track_running_stats = batchnorm_track_running_stats
        self.activation = activation or nn.Identity()
        self.residual_merge = residual_merge
        self.merge_after_conv = merge_after_conv
        self.disable_batchnorm = disable_batchnorm
        self.dtype = dtype

        # inner modules
        self.conv = CausalConv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride_f=stride_f,
            dilation=self.dilation,
            bias=False,
            separable=self.separable,
            dtype=self.dtype,
        )

        if self.disable_batchnorm:
            self.batchnorm = nn.Identity()
        else:
            self.batchnorm = nn.BatchNorm2d(
                num_features=self.out_channels,
                eps=self.batchnorm_eps,
                momentum=self.batchnorm_momentum,
                affine=self.batchnorm_affine,
                track_running_stats=self.batchnorm_track_running_stats,
                dtype=self.dtype,
            )

    def forward(self, x: Tensor) -> Tensor:
        z = x
        x = self.conv(x)
        y = self.batchnorm(x)
        y = self.activation(y)
        if self.residual_merge is not None:
            y = self.residual_merge(x if self.merge_after_conv else z, y)
        return y


class DenseConvBlock(Module):
    def __init__(
        self,
        channels: int,
        kernel_size: OneOrPair[int],
        feature_size: int,
        dilation: Optional[OneOrPair[int]] = None,
        disable_dilation_f: bool = False,
        depth: int = 3,
        final_stride: int = 1,
        batchnorm_eps: float = 1e-05,
        batchnorm_affine: bool = True,
        disable_layernorm: bool = False,
        activation: Module = nn.LeakyReLU(),
        dtype=None,
    ):
        """
        Dense block from Dense CNN With Self-Attention for Time-Domain Speech Enhancement
        paper (https://ieeexplore.ieee.org/document/9372863).

        Parameters
        ----------
        channels : int
            Number of input and output channels
        kernel_size : OneOrPair[int]
            Size of the convolutional kernels
        feature_size : int
            Size of the last dimension
        dilation : Optional[Tuple[int,int]]
            By default the dilation is equal to the kernel to the power of
            the post_conv layer index
        disable_dilation_f : bool
            If True dilation_f==1 for each conv dilation setting,
            by default False
        tconv_kernel_f : Optional[int], optional
            Frequncy kernel size of the transposed convolution,
            by default twice tconv_stride_f
        final_stride : int, optional
            Stride of the last convolution, by default 1
        batchnorm_eps : float, optional
            Eps parameter of the BatchNorm2d , by default 1e-05
        batchnorm_affine : bool, optional
            Affine parameter of the BatchNorm2d, by default True
        disable_layernorm : bool, optional
            Disables the batch normalization, by default False
        activation : Module, optional
            Activation to use, by default nn.LeakyReLU()
        dtype : optional
            Module dtype, by default None
        """
        super().__init__()

        # attributes
        self.channels = channels
        self.kernel_size = kernel_size
        self.feature_size = feature_size
        self.dilation = dilation
        self.disable_dilation_f = disable_dilation_f
        self.depth = depth
        self.final_stride = final_stride
        self.batchnorm_eps = batchnorm_eps
        self.batchnorm_affine = batchnorm_affine
        self.disable_layernorm = disable_layernorm
        self.activation = activation or nn.Identity()
        self.dtype = dtype

        # default dilation ~ ~ ~ ~ ~ ~
        self.dilation = (
            get_default_dilation(self.kernel_size, self.depth, self.disable_dilation_f)
            if self.dilation is None
            else [self.dilation] * self.depth
        )
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

        # inner modules
        _sample_norm = lambda i: (
            nn.Identity()
            if (self.disable_layernorm or i == (self.depth - 1))
            else nn.LayerNorm(
                normalized_shape=self.feature_size,
                eps=self.batchnorm_eps,
                elementwise_affine=self.batchnorm_affine,
                dtype=self.dtype,
            )
        )

        _block = lambda i, k, d: nn.Sequential(
            CausalConv2dNormAct(
                in_channels=(i + 1) * self.channels,
                out_channels=self.channels,
                kernel_size=k,
                stride_f=self.final_stride if i == (self.depth - 1) else 1,
                dilation=d,
                separable=False,
                activation=None,
                disable_batchnorm=True,
                dtype=self.dtype,
            ),
            _sample_norm(i),
            self.activation,
        )

        kernels = [self.kernel_size] * self.depth
        layers = [
            _block(i, k, d) for i, (k, d) in enumerate(zip(kernels, self.dilation))
        ]

        self.layers = nn.Sequential(*layers)

        # concat over channels
        self._concat = lambda x, y: torch.cat((x, y), dim=1)

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Signal of shape (B, C, T, F)

        Returns
        -------
        Tensor
            Output of shape (B, C, T, F//final_stride)
        """
        for i, block in enumerate(self.layers):
            y = block(x)
            x = self._concat(x, y) if i != (self.depth - 1) else y

        return x


# recurrent layers variants = = = = = = = = = = = = = =
class GruNormAct(Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        gru_bias: bool = False,
        batch_first: bool = True,
        dropout: float = 0,
        bidirectional: bool = False,
        eps: float = 1e-05,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
        activation: Module = nn.ReLU(),
        residual_merge: Optional[Callable] = None,
        disable_batchnorm: bool = False,
        dtype=None,
    ):
        """
        GRU + BatchNorm1d + Activation.

        Parameters
        ----------
        Combination of the modules parameters

        activation: Module, optional
            Activation module, by default nn.Relu()
        residual_merge: Optional[Callable], optional
            If different da None, it indicates the merge operation after
            the activation, by default None
        disable_batchnorm: bool, optional
            Disable the BatchNorm2d layer, by default False
        """
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.gru_bias = gru_bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats
        self.activation = activation or nn.Identity()
        self.residual_merge = residual_merge
        self.disable_batchnorm = disable_batchnorm
        self.dtype = dtype

        # inner modules
        self.gru = nn.GRU(
            input_size=self.input_size,
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            bias=self.gru_bias,
            batch_first=self.batch_first,
            dropout=self.dropout,
            bidirectional=self.bidirectional,
            dtype=self.dtype,
        )

        if self.disable_batchnorm:
            self.batchnorm = nn.Identity()
        else:
            self.batchnorm = nn.BatchNorm1d(
                num_features=self.hidden_size,
                eps=self.eps,
                momentum=self.momentum,
                affine=self.affine,
                track_running_stats=self.track_running_stats,
                dtype=self.dtype,
            )

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        y, h = self.gru(x)
        y = y.transpose(1, 2)
        y = self.batchnorm(y)
        y = y.transpose(1, 2)
        y = self.activation(y)
        if self.residual_merge is not None:
            y = self.residual_merge(x, y)
        return y, h


# attention variants  = = = = = = = = = = = = = = = = =


# This is untested
class SlidingCausalMultiheadAttention(Module):
    def __init__(
        self,
        channels: int,
        sequence_len: int,
        embed_dim: int,
        stride: int,
        num_heads: int,
        dropout: float = 0.0,
        bias: bool = True,
        receptive_field: Optional[int] = None,
        attn_mask: Optional[Tensor] = None,
    ):
        """
        Causal multi head attention module applied on an
        unfolded version of the input.

        Parameters
        ----------
        channels : int
            Number of input channels
        embed_dim : int
            Dimension of the embeddings (F)
        sequence_len : int
            Lenght of the sequence blocks resulting from the unfolding
        stride : int
            Hop size between the blocks
        num_heads : int
            Number of heads for the attention
        dropout : float, optional
            Dropout amount, by default 0.0
        bias : bool, optional
            Enables/disables the biases in the projections, by default True
        receptive_field : Optional[int], optional
            Used for the default causal Longformer mask, indicates the number of frames
            employed in the computation of the attention, by default half of the sequence_len
        attn_mask : Optional[Tensor], optional
            Mask provided to the attention layer, by default a causal version of the Longformer
            mask is provided https://arxiv.org/abs/2004.05150
        """
        super().__init__()

        # attributes
        self.channels = channels
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.sequence_len = sequence_len
        self.stride = stride
        self.dropout = dropout
        self.bias = bias
        self.receptive_field = receptive_field or (self.sequence_len // 2)
        self.attn_mask = attn_mask or get_causal_longformer_mask(
            self.sequence_len, self.receptive_field
        )

        # inner modules
        self.reshape_0 = lambda x: x.flatten(0, 2)
        self.reshape_1 = lambda x, b, c: x.reshape(b, c, -1, *x.shape[-2:])

        self.unfold = UnfoldSpectrogram(self.sequence_len, self.stride)
        self.fold = FoldSpectrogram(self.sequence_len, self.stride, self.channels)

        self.mh_attention = nn.MultiheadAttention(
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            dropout=self.dropout,
            bias=self.bias,
            batch_first=True,
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Input of shape (B, C, T, F)

        Returns
        -------
        Tensor
            Output of shape (B, C, T, F)
        """
        b, c = x.shape[:2]
        x = self.unfold(x)
        x = self.reshape_0(x)
        x = self.mh_attention(x, x, x, attn_mask=self.attn_mask)[0]
        x = self.reshape_1(x, b, c)
        x = self.fold(x)
        return x


# = = = = = = = = = = = = = = = = = = = = = = = = = = =



__all__ = [
    "enable_anomaly_detection",
    "get_submodules",
    "compute_gradients",
    "get_num_parameters",
    "freeze_model",
]

from torch import autograd as AG
from torch import nn, Tensor
from typing import Sequence
import torch


def enable_anomaly_detection():
    """
    Enable PyTorch anomaly detection
    """
    from torch import autograd

    autograd.anomaly_mode.set_detect_anomaly(True)


def get_submodules(model: nn.Module) -> Sequence[nn.Module]:
    """
    Gets all the submodules without children from a model.

    Parameters
    ----------
    model : nn.Module
        Target model

    Returns
    -------
    Sequence[nn.Module]
        Model submodules
    """
    modules = [x for x in model.modules() if len(list(x.children())) == 0]
    return modules


def compute_gradients(x: Tensor, y: Tensor, keep_graph: bool = True) -> Tensor:
    """
    Computes the matrix-vector product between the Jacobian calculated from the
    function with y as output and x as input and a vector of ones of the same
    shape as y.

    Parameters
    ----------
    x : Tensor
        Input of the target function, must require grad
    y : Tensor
        Output of the target function
    keep_graph : bool, optional
        If True the result can be used in further autograd graphs, it can then be used
        for higher degree derivative optimization (i.e. as an derivative optimization term in a loss
        such in https://arxiv.org/abs/1704.00028), by default True

    Returns
    -------
    Tensor
        matrix-vector product between the Jacobian
    """
    ones = torch.ones_like(y)
    grad = AG.grad(y, x, ones, retain_graph=keep_graph, create_graph=keep_graph)[0]
    return grad


def get_num_parameters(model: nn.Module) -> int:
    """
    Gets the total number of parameters (trainable and untrainable).

    Parameters
    ----------
    model : nn.Module
        Target model

    Returns
    -------
    int
        Total number of parameters
    """
    return sum(p.numel() for p in model.parameters())


def freeze_model(model: nn.Module):
    """
    Freezed the weights of a target module.

    Parameters
    ----------
    model : nn.Module
        Model to freeze
    """
    for p in model.parameters():
        p.requires_grad = False



from typing import Callable, Optional
from torch import nn, Tensor
from torch.nn import Module
import torch
import pytest


from torch_utilities.utilities import OneOrPair

from torch_utilities.modules import (
    SlidingCausalMultiheadAttention,
    CausalConv2dNormAct,
    UnfoldSpectrogram,
    ScaleChannels2d,
    FoldSpectrogram,
    CausalSubConv2d,
    get_freq_value,
    Reparameterize,
    DenseConvBlock,
    GroupedLinear,
    ResidualWrap,
    CausalConv2d,
    GruNormAct,
    LookAhead,
)


# Local fixtures ===============================================================


@pytest.fixture(params=[2])
def channels(request) -> int:
    """Number of channels."""
    return request.param


@pytest.fixture(params=[True, False])
def lookahead_maintain_shape(request) -> bool:
    """Maintain shape in Lookahead."""
    return request.param


# unfold block sie
@pytest.fixture(params=[2, 4])
def unfold_block_size(request) -> int:
    """Unfold Block size."""
    return request.param


@pytest.fixture(params=[1, 2])
def unfold_stride(request) -> int:
    """Unfold stride."""
    return request.param


@pytest.fixture
def unfold_input(channels) -> Tensor:
    """Unfold input."""
    return torch.randn(1, channels, 10, 16)


@pytest.fixture
def fold_input(unfold_block_size, channels) -> Tensor:
    """Unfold output."""
    return torch.randn(1, channels, 3, unfold_block_size, 16)


@pytest.fixture
def unfold_instance(unfold_block_size, unfold_stride) -> UnfoldSpectrogram:
    """Unfold instance."""
    return UnfoldSpectrogram(unfold_block_size, unfold_stride)


@pytest.fixture
def fold_instance(unfold_block_size, unfold_stride, channels) -> FoldSpectrogram:
    """Fold instance."""
    return FoldSpectrogram(unfold_block_size, unfold_stride, channels)


@pytest.fixture(params=[4, (2, 3)])
def kernel_size(request) -> OneOrPair:
    """Conv kernel size."""
    return request.param


@pytest.fixture(params=[1, 4])
def stride_f(request) -> int:
    """Conv stride over frequencies."""
    return request.param


@pytest.fixture(params=[None, 0, 2])
def padding_f(request) -> Optional[int]:
    """Conv padding over frequencies."""
    return request.param


@pytest.fixture(params=[1, (3, 2)])
def dilation(request) -> OneOrPair:
    """Conv dilation."""
    return request.param


@pytest.fixture(params=[True])
def bias(request) -> bool:
    """Conv bias."""
    return request.param


@pytest.fixture(params=[True, False])
def separable(request) -> bool:
    """Enable separable convolution."""
    return request.param


@pytest.fixture(params=[2])
def in_channels(request) -> int:
    """Layer input channels."""
    return request.param


@pytest.fixture(params=[3])
def out_channels(request) -> int:
    """Layer output channels."""
    return request.param


@pytest.fixture(params=[10])
def n_frames(request) -> int:
    """Number of STFT frames."""
    return request.param


@pytest.fixture(params=[16])
def in_freqs(request) -> int:
    """Number of input frequencies."""
    return request.param


@pytest.fixture(params=[1])
def batch_size(request) -> int:
    """Batch size."""
    return request.param


@pytest.fixture()
def stft_input(in_channels, batch_size, n_frames, in_freqs) -> Tensor:
    """STFT input tensor."""
    return torch.randn(batch_size, in_channels, n_frames, in_freqs)


@pytest.fixture
def causal_conv2d_instance(
    in_channels,
    out_channels,
    kernel_size,
    stride_f,
    padding_f,
    dilation,
    bias,
    separable,
) -> CausalConv2d:
    """CausalConv2d instance."""
    return CausalConv2d(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride_f=stride_f,
        padding_f=padding_f,
        dilation=dilation,
        bias=bias,
        separable=separable,
    )


@pytest.fixture
def causal_conv2d_norm_act_instance(
    in_channels,
    out_channels,
    kernel_size,
    stride_f,
    dilation,
    separable,
) -> CausalConv2dNormAct:
    """CausalConv2dNormAct instance."""
    return CausalConv2dNormAct(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride_f=stride_f,
        dilation=dilation,
        separable=separable,
    )


@pytest.fixture(params=[8])
def hidden_size(request) -> int:
    """Hidden size."""
    return request.param


@pytest.fixture(params=[None, (lambda x, y: y)])
def merge(request) -> Optional[Callable]:
    """Merge operation."""
    return request.param


@pytest.fixture(params=[None, nn.Sigmoid()])
def activation(request) -> Optional[Module]:
    """Activation function."""
    return request.param


@pytest.fixture(params=[True, False])
def enable_batchnorm(request) -> bool:
    """Enable batchnorm."""
    return request.param


@pytest.fixture
def gru_norm_act_instance(
    in_freqs,
    hidden_size,
    merge,
    activation,
    enable_batchnorm,
) -> GruNormAct:
    """GruNormAct instance."""
    return GruNormAct(
        input_size=in_freqs,
        hidden_size=hidden_size,
        residual_merge=merge,
        activation=activation,
        disable_batchnorm=not enable_batchnorm,
    )


@pytest.fixture
def gru_input(in_freqs) -> Tensor:
    """GRU input."""
    return torch.randn(1, 10, in_freqs)


@pytest.fixture(params=[1, 2])
def final_stride(request) -> int:
    """Final stride for DenseConvBlock."""
    return request.param


@pytest.fixture(params=[1, 4])
def depth(request) -> int:
    """Depth of DenseConvBlock."""
    return request.param


@pytest.fixture(params=[False, True])
def disable_dilation_f(request) -> bool:
    """Disable dilation."""
    return request.param


@pytest.fixture
def dense_conv_block_instance(
    channels,
    kernel_size,
    dilation,
    disable_dilation_f,
    depth,
    stride_f,
    activation,
    in_freqs,
    enable_batchnorm,
) -> DenseConvBlock:
    """DenseConvBlock instance."""
    return DenseConvBlock(
        channels=channels,
        kernel_size=kernel_size,
        dilation=dilation,
        disable_dilation_f=disable_dilation_f,
        depth=depth,
        final_stride=stride_f,
        disable_layernorm=not enable_batchnorm,
        activation=activation,
        feature_size=in_freqs,
    )


@pytest.fixture
def causal_sub_conv2d_instance(
    in_channels,
    out_channels,
    kernel_size,
    stride_f,
    dilation,
    separable,
) -> CausalSubConv2d:
    return CausalSubConv2d(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=stride_f,
        dilation=dilation,
        bias=False,
        separable=separable,
    )


@pytest.fixture
def sliding_causal_multihead_attention_instance(
    stride_f, in_freqs
) -> SlidingCausalMultiheadAttention:
    return SlidingCausalMultiheadAttention(
        channels=2,
        sequence_len=10,
        embed_dim=in_freqs,
        stride=stride_f,
        num_heads=2,
        dropout=0,
        bias=False,
        receptive_field=None,
        attn_mask=None,
    )


# ==============================================================================


class TestLookAhead:
    x = torch.zeros(1, 2, 10, 16)

    def test_forward_shape(self, lookahead_maintain_shape):
        expected_shape = (1, 2, 6, 16) if lookahead_maintain_shape else (1, 2, 6, 16)
        lookahead = LookAhead(4)
        y = lookahead(self.x)
        assert y.shape == expected_shape


class TestUnfoldFoldSpectrogram:
    def test_unfold_forward_shape(
        self, unfold_instance, unfold_input, unfold_block_size, unfold_stride
    ):
        batch_size, channels, frames, freqs = unfold_input.shape
        unfold = unfold_instance
        x = unfold_input
        y = unfold(x)

        n_blocks = (frames - (unfold_block_size - 1) - 1) / unfold_stride + 1
        expected_shape = (
            batch_size,
            channels,
            n_blocks,
            unfold_block_size,
            freqs,
        )
        assert y.shape == expected_shape

    def test_fold_forward_shape(self, fold_instance, fold_input):
        batch_size, channels, blocks, block_size, freqs = fold_input.shape
        fold = fold_instance
        x = fold_input
        y = fold(x)
        assert y.shape[:2] == (batch_size, channels)
        assert y.shape[3] == freqs

    def test_inversion(self, fold_instance, unfold_instance, unfold_input):
        unfold = unfold_instance
        fold = fold_instance
        x = unfold_input
        y = unfold(x)
        x_hat = fold(y)
        torch.testing.assert_close(x, x_hat)


class TestResidualWrap:
    x = torch.ones(1, 2, 10, 16)

    def test_single_layer(self):
        layer = ResidualWrap(nn.Identity())
        y = layer(self.x)
        assert y.shape == self.x.shape
        torch.testing.assert_close(2 * self.x, y)

    def test_multiple_layers(self):
        layer = ResidualWrap(nn.Identity(), nn.Identity())
        y = layer(self.x)
        assert y.shape == self.x.shape
        torch.testing.assert_close(2 * self.x, y)


class TestCausalConv2d:
    def test_forward_shape(
        self,
        causal_conv2d_instance,
        padding_f,
        stride_f,
        in_freqs,
        stft_input,
        dilation,
        batch_size,
        out_channels,
        kernel_size,
        n_frames,
    ):
        if padding_f is None and in_freqs % stride_f != 0:
            pytest.skip(
                "Skipping autopadding test when stride_f doesn't divide in_freqs"
            )

        conv = causal_conv2d_instance
        x = stft_input
        y = conv(x)

        dilation_f = get_freq_value(dilation)
        kernel_f = get_freq_value(kernel_size)

        if padding_f is None:
            # auto padding
            out_freqs = in_freqs // stride_f
        else:
            # manual padding
            out_freqs = in_freqs + 2 * padding_f - dilation_f * (kernel_f - 1) - 1
            out_freqs = int(out_freqs / stride_f + 1)

        expected_shape = (batch_size, out_channels, n_frames, out_freqs)
        assert y.shape == expected_shape


class TestCausalConv2dNormAct:
    def test_forward_shape(
        self,
        causal_conv2d_norm_act_instance,
        stride_f,
        in_freqs,
        stft_input,
        batch_size,
        out_channels,
        n_frames,
    ):
        conv = causal_conv2d_norm_act_instance
        x = stft_input
        y = conv(x)
        out_freqs = in_freqs // stride_f
        expected_shape = (batch_size, out_channels, n_frames, out_freqs)
        assert y.shape == expected_shape


class TestReparameterize:
    reparam = Reparameterize()
    eps = 1e-1

    def test_zero_mean(self):
        mu = torch.zeros(100000)
        logvar = torch.ones(100000)
        y = self.reparam(mu, logvar)
        y = torch.mean(y)
        assert y.item() < self.eps

    def test_nonzero_mean(self):
        mu = torch.ones(100000)
        logvar = torch.ones(100000)
        y = self.reparam(mu, logvar)
        y = torch.mean(y) - 1
        assert y.item() < self.eps


class TestScaleChannels2d:
    scale = ScaleChannels2d(2)

    def set_weights(self, weights: list):
        w = self.scale.scale.weight.data
        self.scale.scale.weight.data = torch.ones_like(w)
        for i, w in enumerate(weights):
            self.scale.scale.weight.data[i] = w

    def test_scaling(self):
        self.set_weights([1, 2])
        x = torch.ones((1, 2, 10, 32))
        y = self.scale(x)
        torch.testing.assert_close(y[0, 0], x[0, 0])
        torch.testing.assert_close(y[0, 1], x[0, 1] * 2)


class TestGroupedLinear:
    def test_assert(self):
        with pytest.raises(AssertionError):
            GroupedLinear(64, 11, groups=8)

    def test_no_groups(self):
        x = torch.rand((1, 10, 32))
        gl = GroupedLinear(32, 64, 1)
        lin = nn.Linear(32, 64, bias=False)
        gl.weight.data = torch.ones_like(gl.weight.data)
        lin.weight.data = torch.ones_like(lin.weight.data)
        torch.testing.assert_close(gl(x), lin(x))


class TestGruNormAct:
    def test_forward_shape(self, gru_norm_act_instance, gru_input, hidden_size):
        x = gru_input
        gru = gru_norm_act_instance
        y, h = gru(x)
        assert y.shape == (*x.shape[:-1], hidden_size)
        assert h.shape == (x.shape[0], 1, hidden_size)


class TestDenseConvBlock:
    def test_forward_shape(self, stft_input, dense_conv_block_instance, stride_f):
        dcb = dense_conv_block_instance
        x = stft_input
        y = dcb(x)
        B, C, T, F = x.shape
        assert y.shape == (B, C, T, F // stride_f)


class TestCausalSubConv2d:
    def test_forward(
        self,
        causal_sub_conv2d_instance,
        stft_input,
        out_channels,
        stride_f,
    ):
        csc = causal_sub_conv2d_instance
        x = stft_input
        y = csc(x)
        B, C, T, F = x.shape
        expected_shape = (B, out_channels, T, F * stride_f)
        assert y.shape == expected_shape


class TestSlidingCausalMultiheadAttention:
    def test_forward_shape(
        self, sliding_causal_multihead_attention_instance, stft_input
    ):
        att = sliding_causal_multihead_attention_instance
        x = stft_input
        y = att(x)
        assert x.shape == y.shape



from torch import nn
import pytest
import torch

import torch_utilities as tu

# Local fixtures ===============================================================


@pytest.fixture(params=[False, True])
def keep_graph(request) -> bool:
    """Keep graph for the gradient computation."""
    return request.param


# ==============================================================================


class TestPyTorch:
    @torch.set_grad_enabled(True)
    def test_compute_gradients(self, keep_graph):
        x = torch.ones((3, 2), requires_grad=True)
        y = x**2
        grad = tu.compute_gradients(x, y, keep_graph=keep_graph)
        z = grad.sum()
        assert z.item() <= 2 * 6
        if keep_graph:
            z.backward()
        else:
            with pytest.raises(RuntimeError):
                z.backward()

    def test_get_submodules(self):
        model = nn.Sequential(
            nn.Identity(),
            nn.ReLU(),
            nn.Tanh(),
            nn.Sequential(
                nn.SELU(),
                nn.Sigmoid(),
            ),
        )
        modules = tu.get_submodules(model)
        modules_types = [type(m) for m in modules]
        expected = [nn.Identity, nn.ReLU, nn.Tanh, nn.SELU, nn.Sigmoid]
        assert modules_types == expected

    @torch.set_grad_enabled(True)
    def test_freeze_model(self):
        module = nn.Conv2d(2, 2, 1)
        assert all(p.requires_grad for p in module.parameters())
        tu.freeze_model(module)
        assert not all(p.requires_grad for p in module.parameters())



from functools import reduce
from operator import mul
from typing import List
import numpy as np
import pytest
import torch

import torch_utilities as tu

# Local fixtures ===============================================================


# expected factorizations
@pytest.fixture(params=[[3, 7, 11], [3, 17], [2, 2, 2, 7]])
def expected_factorization(request) -> List[int]:
    return request.param


# ==============================================================================


class TestUtilities:
    def test_pack_many(self):
        xs = [1, 2, 3]
        ys = [4, 5, 6]
        zss = tu.pack_many(xs, ys)
        assert zss == [(1, 4), (2, 5), (3, 6)]

    def test_phase(self, module):
        phase = np.array([0, 1, 0.2])
        phase = np.exp(1j * phase)
        if module == torch:
            phase = torch.from_numpy(phase)
        x = module.ones_like(phase)
        x = x * phase
        phase_hat = tu.phase(x)
        torch.testing.assert_close(phase, phase_hat)

    def test_factorize(self, expected_factorization):
        exp = expected_factorization
        x = reduce(mul, exp)
        y = tu.factorize(x)
        assert y == exp
